from xdsl.ir import (
    Dialect,
    ParametrizedAttribute,
    TypeAttribute,
)
from xdsl.irdl import (
    irdl_attr_definition,
)


@irdl_attr_definition
class AffineMapType(ParametrizedAttribute, TypeAttribute):
    name = "transform.affine_map"


@irdl_attr_definition
class AnyOpType(ParametrizedAttribute, TypeAttribute):
    name = "transform.any_op"


@irdl_attr_definition
class AnyParamType(ParametrizedAttribute, TypeAttribute):
    name = "transform.any_param"


@irdl_attr_definition
class AnyValueType(ParametrizedAttribute, TypeAttribute):
    name = "transform.any_value"


Transform = Dialect(
    "transform",
    [],
    [
        # TYpes
        AffineMapType,
        AnyOpType,
        AnyParamType,
        AnyValueType,
    ],
)



from collections.abc import Sequence
from dataclasses import dataclass

from xdsl.builder import ImplicitBuilder
from xdsl.context import MLContext
from xdsl.dialects import arith, builtin, func, stencil
from xdsl.dialects.builtin import IntegerAttr, TensorType
from xdsl.dialects.csl import csl, csl_stencil, csl_wrapper
from xdsl.ir import Attribute
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.hints import isa


@dataclass(frozen=True)
class ConvertStencilFuncToModuleWrappedPattern(RewritePattern):
    """
    Wraps program in the csl_stencil dialect in a csl_wrapper module.
    Scans the csl_stencil.apply ops for stencil-related params, passing them as properties to the wrapped module
    (note, properties are in return passed as block args to the layout_module and program_module blocks).

    The layout module wrapper can be used to initialise general program module params. This pass generates code
    to initialise stencil-specific program params and yields them from the layout module.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter, /):
        # find csl_stencil.apply ops, abort if there are none
        apply_ops = self.get_csl_stencil_apply_ops(op)
        if len(apply_ops) == 0:
            return
        max_distance: int = 1
        width: int = 1
        height: int = 1
        z_dim_no_ghost_cells: int = 1
        z_dim: int = 1
        for apply_op in apply_ops:
            # loop over accesses to get max_distance (from which we build `pattern`)
            for ap in apply_op.get_accesses():
                if ap.is_diagonal:
                    raise ValueError("Diagonal accesses are currently not supported")
                if len(ap.offsets) > 0:
                    if ap.dims != 2:
                        raise ValueError(
                            "Stencil accesses must be 2-dimensional at this stage"
                        )
                    max_distance = max(max_distance, ap.max_distance())

            # find max x and y dimensions
            if len(shape := apply_op.topo.shape.data) == 2:
                assert isinstance(
                    shape.data[0].data, int
                ), "Cannot have a float data shape"
                assert isinstance(
                    shape.data[1].data, int
                ), "Cannot have a float data shape"
                width = max(width, shape.data[0].data)
                height = max(height, shape.data[1].data)
            else:
                raise ValueError("Stencil accesses must be 2-dimensional at this stage")

            # find max z dimension - we could get this from func args, store ops, or apply ops
            for result in apply_op.results:
                if isa(result.type, stencil.TempType[TensorType[Attribute]]):
                    z_dim_no_ghost_cells = max(
                        z_dim_no_ghost_cells,
                        result.type.get_element_type().get_shape()[0],
                    )
            for arg in op.args:
                if isa(field_t := arg.type, stencil.FieldType[TensorType[Attribute]]):
                    z_dim = max(z_dim, field_t.get_element_type().get_shape()[0])

        # initialise module op
        module_op = csl_wrapper.ModuleOp(
            width=IntegerAttr(width, 16),
            height=IntegerAttr(height, 16),
            params={
                "z_dim": IntegerAttr(z_dim, 16),
                "pattern": IntegerAttr(max_distance + 1, 16),
            },
        )

        self.initialise_layout_module(module_op)
        module_op.program_name = op.sym_name

        # add yield op args (implicit) and func args (explicit) to program_module block args
        module_op.update_program_block_args(
            exported_symbols=[(arg.name_hint, arg.type) for arg in op.body.block.args]
        )

        # replace func.return
        func_return = op.body.block.last_op
        assert isinstance(func_return, func.Return)
        assert (
            len(func_return.arguments) == 0
        ), "Non-empty returns currently not supported"
        rewriter.replace_op(func_return, csl.ReturnOp())

        # set up main function and move func.func ops into this csl.func
        main_func = csl.FuncOp(op.sym_name.data, ((), None))
        rewriter.inline_block(
            op.body.block,
            InsertPoint.at_start(main_func.body.block),
            module_op.exported_symbols,
        )

        # add main and empty yield to program_module
        module_op.program_module.block.add_ops([main_func, csl_wrapper.YieldOp([], [])])

        # replace (now empty) func by module wrapper
        rewriter.replace_matched_op(module_op)

    def get_csl_stencil_apply_ops(
        self, op: func.FuncOp
    ) -> Sequence[csl_stencil.ApplyOp]:
        result: list[csl_stencil.ApplyOp] = []
        for apply_op in op.body.walk():
            if isinstance(apply_op, csl_stencil.ApplyOp):
                result.append(apply_op)
        return result

    def initialise_layout_module(self, module_op: csl_wrapper.ModuleOp):
        """Initialises the layout_module (wrapper block) by setting up (esp. stencil-related) program params"""

        # extract layout module params as the function has linear complexity
        param_width = module_op.get_layout_param("width")
        param_height = module_op.get_layout_param("height")
        param_x = module_op.get_layout_param("x")
        param_y = module_op.get_layout_param("y")
        param_pattern = module_op.get_layout_param("pattern")

        # fill layout module wrapper block with ops
        with ImplicitBuilder(module_op.layout_module.block):
            # set up LAUNCH
            zero = arith.Constant(IntegerAttr(0, 16))
            launch = csl.GetColorOp(zero)

            # import memcpy/get_params and routes
            memcpy = csl_wrapper.ImportOp(
                "<memcpy/get_params>",
                {
                    "width": param_width,
                    "height": param_height,
                    "LAUNCH": launch.res,
                },
            )
            routes = csl_wrapper.ImportOp(
                "routes.csl",
                {
                    "pattern": param_pattern,
                    "peWidth": param_width,
                    "peHeight": param_height,
                },
            )

            # set up program param `stencil_comms_params`
            all_routes = csl.MemberCallOp(
                "computeAllRoutes",
                csl.ComptimeStructType(),
                routes,
                params=[
                    param_x,
                    param_y,
                    param_width,
                    param_height,
                    param_pattern,
                ],
            )
            # set up program param `memcpy_params`
            memcpy_params = csl.MemberCallOp(
                "get_params",
                csl.ComptimeStructType(),
                memcpy,
                params=[
                    param_x,
                ],
            )

            # set up program param `is_border_region_pe`
            one = arith.Constant(IntegerAttr(1, 16))
            pattern_minus_one = arith.Subi(one, param_pattern)
            width_minus_x = arith.Subi(param_width, param_x)
            height_minus_y = arith.Subi(param_height, param_y)
            x_lt_pattern_minus_one = arith.Cmpi(param_x, pattern_minus_one, "slt")
            y_lt_pattern_minus_one = arith.Cmpi(param_y, pattern_minus_one, "slt")
            width_minus_one_lt_pattern = arith.Cmpi(width_minus_x, param_pattern, "slt")
            height_minus_one_lt_pattern = arith.Cmpi(
                height_minus_y, param_pattern, "slt"
            )
            or1_op = arith.OrI(x_lt_pattern_minus_one, y_lt_pattern_minus_one)
            or2_op = arith.OrI(or1_op, width_minus_one_lt_pattern)
            is_border_region_pe = arith.OrI(or2_op, height_minus_one_lt_pattern)

            # yield things as named params to the program module
            csl_wrapper.YieldOp.from_field_name_mapping(
                field_name_mapping={
                    "memcpy_params": memcpy_params.results[0],
                    "stencil_comms_params": all_routes.results[0],
                    "isBorderRegionPE": is_border_region_pe.result,
                }
            )


@dataclass(frozen=True)
class CslStencilToCslWrapperPass(ModulePass):
    """
    Wraps program in the csl_stencil dialect in a csl_wrapper by translating each
    top-level function to one module wrapper.
    """

    name = "csl-stencil-to-csl-wrapper"

    def apply(self, ctx: MLContext, op: builtin.ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ConvertStencilFuncToModuleWrappedPattern(),
                ]
            ),
            apply_recursively=False,
        )
        module_pass.rewrite_module(op)



from typing import Any, TypeGuard

from xdsl.ir import AttributeInvT
from xdsl.irdl import GenericAttrConstraint
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa


def isattr(
    arg: Any, hint: type[AttributeInvT] | GenericAttrConstraint[AttributeInvT]
) -> TypeGuard[AttributeInvT]:
    from xdsl.irdl import ConstraintContext

    if isinstance(hint, GenericAttrConstraint):
        try:
            hint.verify(arg, ConstraintContext())
            return True
        except VerifyException:
            return False

    return isa(arg, hint)



from collections.abc import Sequence
from dataclasses import dataclass
from math import prod

from xdsl.context import MLContext
from xdsl.dialects import arith, memref, stencil, tensor
from xdsl.dialects.builtin import (
    AnyMemRefType,
    AnyTensorType,
    IndexType,
    IntegerAttr,
    IntegerType,
    ModuleOp,
    TensorType,
)
from xdsl.dialects.csl import csl_stencil
from xdsl.dialects.experimental import dmp
from xdsl.ir import Attribute, Block, BlockArgument, Operation, OpResult, Region
from xdsl.irdl import Operand, base
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.transforms.experimental.stencil_tensorize_z_dimension import (
    BackpropagateStencilShapes,
)
from xdsl.utils.hints import isa
from xdsl.utils.isattr import isattr


def get_stencil_access_operands(op: Operand) -> set[Operand]:
    """
    Returns the symbols of all stencil accessess by op and all its dependencies.
    """
    res: set[Operand] = set()
    frontier: set[Operand] = {op}
    next: set[Operand] = set()
    while len(frontier) > 0:
        for o in frontier:
            if isinstance(o, OpResult):
                if isinstance(o.op, csl_stencil.AccessOp):
                    res.add(o.op.op)
                else:
                    next.update(o.op.operands)
        frontier = next
        next = set()

    return res


@dataclass(frozen=True)
class RestructureSymmetricReductionPattern(RewritePattern):
    """
    Consume data first where that data comes from stencil accesses to `buf`.

    Identifies a pattern of 2 connected binary ops with 3 args, e.g. of the form `(a+b)+c` with different ops and
    bracketings supported, and attempts to re-structure the order of computation.

    Being in principle similarly to constant folding, the difference is that args are *not* required to be stencil
    accesses, but could have further compute applied before being passed to the reduction function.
    Uses helper function `get_stencil_accessed_symbols` to check which bufs are stencil-accessed in each of these args,
    and to distinguish the following three cases:

     (1) all accesses in an arg tree are to `buf`  - arg should be moved forward in the computation
     (2) no accesses are to `buf`                  - arg should be moved backward in the computation
     (3) there's a mix                             - unknown, take any or no action

    If two args are identified that should be moved forward, or two args are identified that should be moved backwards,
    the computation is restructured accordingly.
    """

    buf: BlockArgument

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.Addf | arith.Mulf, rewriter: PatternRewriter, /
    ):
        # this rewrite requires exactly 1 use which is the same type of operation
        if len(op.result.uses) != 1 or not isinstance(
            use := list(op.result.uses)[0].operation, type(op)
        ):
            return
        c_op = use.operands[0] if use.operands[1] == op.result else use.operands[1]

        def rewrite(one: Operand, two: Operand, three: Operand):
            """Builds `(one+two)+three` where `'+' == type(op)`"""

            first_compute = type(op)(one, two)
            second_compute = type(op)(first_compute, three)

            # Both ops are inserted at the later point to ensure all dependencies are present when moving compute around.
            # Moving the replacement of `op` backwards is safe because we previously asserted at `op` only has one use (ie. in `use`)
            rewriter.replace_op(op, [], [first_compute.results[0]])
            rewriter.replace_op(
                use, [first_compute, second_compute], [second_compute.results[0]]
            )

        a = get_stencil_access_operands(a_op := op.lhs)
        b = get_stencil_access_operands(b_op := op.rhs)
        c = get_stencil_access_operands(c_op)

        if self.move_fwd(a) and self.move_fwd(b):
            return
        elif self.move_back(a) and self.move_back(b):
            return
        elif self.move_fwd(c) and self.move_back(b):
            rewrite(a_op, c_op, b_op)

    def move_fwd(self, accs: set[Operand]) -> bool:
        return self.buf in accs and len(accs) == 1

    def move_back(self, accs: set[Operand]) -> bool:
        return self.buf not in accs


@dataclass(frozen=True)
class ConvertAccessOpFromPrefetchPattern(RewritePattern):
    """
    Rebuilds stencil.access by csl_stencil.access which operates on prefetched accesses.

    stencil.access operates on stencil.temp types found at arg_index
    csl_stencil.access operates on memref< num_neighbors x tensor< buf_size x data_type >> found at last arg index

    Note: This is intended to be called in a nested pattern rewriter, such that the above precondition is met.
    """

    arg_index: int

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.AccessOp, rewriter: PatternRewriter, /):
        assert len(op.offset) == 2
        if op.temp != op.get_apply().region.block.args[self.arg_index]:
            return

        # translate access to own data, which operates on stencil.TempType
        if tuple(op.offset) == (0, 0):
            assert isattr(op.res.type, base(AnyTensorType))
            rewriter.replace_matched_op(
                csl_stencil.AccessOp(
                    op=op.temp,
                    offset=op.offset,
                    offset_mapping=op.offset_mapping,
                    result_type=op.res.type,
                )
            )
            return

        prefetched_arg = op.get_apply().region.block.args[-1]
        assert isa(m_type := prefetched_arg.type, memref.MemRefType[Attribute])
        assert isa(t_type := m_type.get_element_type(), TensorType[Attribute])

        csl_access_op = csl_stencil.AccessOp(
            op=prefetched_arg,
            offset=op.offset,
            offset_mapping=op.offset_mapping,
            result_type=t_type,
        )

        # The stencil-tensorize-z-dimension pass inserts tensor.ExtractSliceOps after stencil.access to remove ghost cells.
        # Since ghost cells are not prefetched, these ops can be removed again. Check if the ExtractSliceOp
        # has no other effect and if so, remove both.
        if (
            len(op.res.uses) == 1
            and isinstance(use := list(op.res.uses)[0].operation, tensor.ExtractSliceOp)
            and tuple(d.data for d in use.static_sizes.data) == t_type.get_shape()
            and tuple(d.data for d in use.static_offsets.data) == (0,)
            and tuple(d.data for d in use.static_strides.data) == (1,)
            and len(use.offsets) == 0
            and len(use.sizes) == 0
            and len(use.strides) == 0
        ):
            rewriter.replace_op(use, csl_access_op)
            rewriter.erase_op(op)
        else:
            rewriter.replace_matched_op(csl_access_op)


@dataclass(frozen=True)
class ConvertSwapToPrefetchPattern(RewritePattern):
    """
    Translates dmp.swap to csl_stencil.prefetch
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
        # remove op if it contains no swaps
        if op.swaps is None or len(op.swaps) == 0:
            rewriter.erase_matched_op(False)
            return

        assert all(
            len(swap.size) == 3 for swap in op.swaps
        ), "currently only 3-dimensional stencils are supported"

        assert all(
            swap.size[:2] == (1, 1) for swap in op.swaps
        ), "invoke dmp to decompose from (x,y,z) to (1,1,z)"

        # check that size is uniform
        uniform_size = op.swaps.data[0].size[2]
        assert all(
            swap.size[2] == uniform_size for swap in op.swaps
        ), "all swaps need to be of uniform size"

        assert isinstance(op.input_stencil, OpResult)
        assert isattr(
            op.input_stencil.type,
            base(AnyMemRefType) | base(stencil.AnyTempType),
        )
        assert isa(
            t_type := op.input_stencil.type.get_element_type(), TensorType[Attribute]
        )
        assert op.topo is not None, f"topology on {type(op)} is not given"

        # when translating swaps, remove third dimension
        prefetch_op = csl_stencil.PrefetchOp(
            input_stencil=op.input_stencil.op,
            topo=op.topo,
            swaps=[
                csl_stencil.ExchangeDeclarationAttr(swap.neighbor[:2])
                for swap in op.swaps
            ],
            result_type=memref.MemRefType(
                TensorType(t_type.get_element_type(), (uniform_size,)),
                (len(op.swaps),),
            ),
        )

        # a little hack to get around a check that prevents replacing a no-results op with an n-results op
        rewriter.replace_matched_op(prefetch_op, new_results=[])

        # uses have to be retrieved *before* the loop because of the rewriting happening inside the loop
        uses = list(op.input_stencil.uses)

        # csl_stencil.prefetch, unlike dmp.swap, has a return value. This is added as the last arg
        # to stencil.apply, before rebuilding the op and replacing stencil.access ops by csl_stencil.access ops
        # that reference the prefetched buffers (note, this is only done for neighbor accesses)
        for use in uses:
            if not isinstance(use.operation, stencil.ApplyOp):
                continue
            apply_op = use.operation

            # arg_idx points to the stencil.temp type whose data is prefetched in a separate buffer
            arg_idx = apply_op.args.index(op.input_stencil)

            # add the prefetched buffer as the last arg to stencil.access
            apply_op.region.block.insert_arg(
                prefetch_op.result.type, len(apply_op.args)
            )

            # rebuild stencil.apply op
            r_types = [r.type for r in apply_op.results]
            assert isa(r_types, Sequence[stencil.TempType[Attribute]])
            new_apply_op = stencil.ApplyOp.get(
                [*apply_op.args, prefetch_op.result],
                apply_op.detach_region(apply_op.region),
                r_types,
            )
            rewriter.replace_op(apply_op, new_apply_op)

            # replace stencil.access (operating on stencil.temp at arg_index)
            # with csl_stencil.access (operating on memref at last arg index)
            nested_rewriter = PatternRewriteWalker(
                ConvertAccessOpFromPrefetchPattern(arg_idx)
            )

            nested_rewriter.rewrite_op(new_apply_op)


def get_op_split(
    ops: Sequence[Operation], buf: BlockArgument
) -> tuple[Sequence[Operation], Sequence[Operation]]:
    """
    Returns a split of `ops` into an `(a,b)` tuple, such that:

    - `a` contains neighbour accesses to `buf` plus the minumum set of instructions to reduce the accessed data to 1 thing
    - `b` contains everything else

    If no valid split can be found, return `(ops, [])`.

    This function does not attempt to arithmetically re-structure the computation to obtain a good split. To do this,
    `RestructureSymmetricReductionPattern()` may be executed first.
    """
    a: list[Operation] = []
    b: list[Operation] = []
    rem: list[Operation] = []
    for op in ops:
        if isinstance(op, csl_stencil.AccessOp):
            (b, a)[op.op == buf].append(op)
        else:
            rem.append(op)

    # loop until we can make no more changes, or until only 1 thing computed in `a` is used outside of it
    has_changes = True
    while (
        len(
            # ops in `a` whose results are used outside of `a`
            a_exports := set(
                op
                for op in a
                for result in op.results
                for use in result.uses
                if use.operation not in a
            )
        )
        > 1
        and has_changes
    ):
        has_changes = False

        # find ops that directly depend on `a` but are not themselves in `a`
        for exp in a_exports:
            for result in exp.results:
                for use in result.uses:
                    # op is only movable if *all* operands are already in `a` (and it hasn't been moved yet)
                    if (op := use.operation) in rem and all(
                        x.op in a for x in op.operands if isinstance(x, OpResult)
                    ):
                        has_changes = True
                        a.append(use.operation)
                        rem.remove(use.operation)

    if len(a_exports) == 1:
        return a, b + rem

    # fallback
    # always place `stencil.return` in second block
    return ops[:-1], [ops[-1]]


@dataclass(frozen=True)
class ConvertApplyOpPattern(RewritePattern):
    """
    Fuses a `csl_stencil.prefetch` and a `stencil.apply` to build a `csl_stencil.apply`.

    If there are several candidate prefetch ops, the one with the largest result buffer size is selected.
    The selection is greedy, and could in the future be expanded into a more global selection optimising for minimal
    prefetch overhead across multiple apply ops.

    args:
        num_chunks - number of chunks into which communication and computation should be split.
                     Effectively, the number of times `csl_stencil.apply.chunk_reduce` will be executed and the
                     tensor sizes it handles. Higher values may increase compute overhead but reduce size of
                     communication buffers when lowered.
    """

    num_chunks: int = 1

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.ApplyOp, rewriter: PatternRewriter, /):
        # calculate memory cost of all prefetch operands
        def get_prefetch_overhead(o: OpResult):
            assert isa(o.type, memref.MemRefType[Attribute])
            assert isa(t_type := o.type.get_element_type(), TensorType[Attribute])
            buf_count = prod(o.type.get_shape())
            buf_size = prod(t_type.get_shape())
            return buf_count * buf_size

        candidate_prefetches = [
            (get_prefetch_overhead(o), o)
            for o in op.operands
            if isinstance(o, OpResult) and isinstance(o.op, csl_stencil.PrefetchOp)
        ]
        if len(candidate_prefetches) == 0:
            return

        # select the prefetch with the biggest communication overhead to be fused with matched stencil.apply
        prefetch = max(candidate_prefetches)[1]
        prefetch_idx = op.operands.index(prefetch)
        assert isinstance(prefetch.op, csl_stencil.PrefetchOp)
        communicated_stencil_idx = op.operands.index(prefetch.op.input_stencil)
        assert isinstance(prefetch.op, csl_stencil.PrefetchOp)
        assert isa(prefetch.type, memref.MemRefType[Attribute])
        assert isa(
            prefetch_t_type := prefetch.type.get_element_type(), TensorType[Attribute]
        )
        communicated_stencil_op_arg = prefetch.op.input_stencil

        # add empty tensor before op to be used as `iter_arg`
        # this could potentially be re-used if we have one of the same size lying around
        iter_arg = tensor.EmptyOp((), prefetch.type.get_element_type())
        rewriter.insert_op(iter_arg, InsertPoint.before(op))

        # run pass (on this apply's region only) to consume data from `prefetch` accesses first
        nested_rewriter = PatternRewriteWalker(
            RestructureSymmetricReductionPattern(op.region.block.args[prefetch_idx]),
            walk_reverse=True,
        )
        nested_rewriter.rewrite_op(op)

        # determine how ops should be split across the two regions
        chunk_reduce_ops, post_process_ops = get_op_split(
            list(op.region.block.ops), op.region.block.args[prefetch_idx]
        )

        # fetch what chunk_reduce is computing for
        if isinstance(chunk_reduce_ops[-1], stencil.ReturnOp):
            chunk_res = chunk_reduce_ops[-1].operands[0]
        else:
            chunk_res = chunk_reduce_ops[-1].results[0]

        # after region split, check which block args (from the old ops block) are being accessed in each of the new regions
        # ignore accesses block args which already are part of the region's required signature
        chunk_reduce_used_block_args = sorted(
            set(
                x
                for o in chunk_reduce_ops
                for x in o.operands
                if isinstance(x, BlockArgument) and x.index != prefetch_idx
            ),
            key=lambda b: b.index,
        )
        post_process_used_block_args = sorted(
            set(
                x
                for o in post_process_ops
                for x in o.operands
                if isinstance(x, BlockArgument) and x.index != communicated_stencil_idx
            ),
            key=lambda b: b.index,
        )

        # set up region signatures, comprising fixed and optional args - see docs on `csl_stencil.apply` for details
        chunk_reduce_args = [
            # required arg 0: slice of type(%prefetch)
            memref.MemRefType(
                TensorType(
                    prefetch_t_type.get_element_type(),
                    (prefetch_t_type.get_shape()[0] // self.num_chunks,),
                ),
                (len(prefetch.op.swaps),),
            ),
            # required arg 1: %offset
            IndexType(),
            # required arg 2: %iter_arg
            iter_arg.results[0].type,
            # optional args: as needed by the ops
            *[a.type for a in chunk_reduce_used_block_args],
        ]
        post_process_args = [
            # required arg 0: stencil.temp to access own data
            communicated_stencil_op_arg.type,
            # required arg 1: %iter_arg
            iter_arg.results[0].type,
            # optional args: as needed by the ops
            *[a.type for a in post_process_used_block_args],
        ]

        # set up two regions
        chunk_reduce = Region(Block(arg_types=chunk_reduce_args))
        post_process = Region(Block(arg_types=post_process_args))

        # translate old to new block arg index for optional args
        chunk_reduce_oprnd_table = dict[Operand, Operand](
            (old, chunk_reduce.block.args[idx])
            for idx, old in enumerate(chunk_reduce_used_block_args, start=3)
        )
        post_process_oprnd_table = dict[Operand, Operand](
            (old, post_process.block.args[idx])
            for idx, old in enumerate(post_process_used_block_args, start=2)
        )

        # add translation from old to new arg index for non-optional args - note, access to iter_arg must be handled separately below
        chunk_reduce_oprnd_table[op.region.block.args[prefetch_idx]] = (
            chunk_reduce.block.args[0]
        )
        post_process_oprnd_table[op.region.block.args[communicated_stencil_idx]] = (
            post_process.block.args[0]
        )
        post_process_oprnd_table[chunk_res] = post_process.block.args[1]

        # detach ops from old region
        for o in op.region.block.ops:
            op.region.block.detach_op(o)

        # add operations from list to chunk_reduce, use translation table to rebuild operands
        for o in chunk_reduce_ops:
            if isinstance(o, stencil.ReturnOp | csl_stencil.YieldOp):
                break
            o.operands = [chunk_reduce_oprnd_table.get(x, x) for x in o.operands]
            chunk_reduce.block.add_op(o)

        # put `chunk_res` into `iter_arg` (using tensor.insert_slice) and yield the result
        chunk_reduce.block.add_ops(
            [
                insert_slice_op := tensor.InsertSliceOp.get(
                    source=chunk_res,
                    dest=chunk_reduce.block.args[2],
                    offsets=(chunk_reduce.block.args[1],),
                    static_sizes=(prefetch_t_type.get_shape()[0] // self.num_chunks,),
                ),
                csl_stencil.YieldOp(insert_slice_op.result),
            ]
        )

        # add operations from list to post_process, use translation table to rebuild operands
        for o in post_process_ops:
            o.operands = [post_process_oprnd_table.get(x, x) for x in o.operands]
            post_process.block.add_op(o)
            if isinstance(o, stencil.ReturnOp):
                rewriter.replace_op(o, csl_stencil.YieldOp(*o.operands))

        rewriter.replace_matched_op(
            csl_stencil.ApplyOp(
                operands=[
                    communicated_stencil_op_arg,
                    iter_arg,
                    [op.operands[a.index] for a in chunk_reduce_used_block_args]
                    + [op.operands[a.index] for a in post_process_used_block_args],
                ],
                properties={
                    "swaps": prefetch.op.swaps,
                    "topo": prefetch.op.topo,
                    "num_chunks": IntegerAttr(self.num_chunks, IntegerType(64)),
                },
                regions=[
                    chunk_reduce,
                    post_process,
                ],
                result_types=[r.type for r in op.results],
            )
        )

        if len(prefetch.uses) == 0:
            rewriter.erase_op(prefetch.op)


@dataclass(frozen=True)
class StencilToCslStencilPass(ModulePass):
    name = "stencil-to-csl-stencil"

    # chunks into which to slice communication
    num_chunks: int = 1

    def apply(self, ctx: MLContext, op: ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ConvertSwapToPrefetchPattern(),
                    ConvertApplyOpPattern(num_chunks=self.num_chunks),
                ]
            ),
            walk_reverse=False,
            apply_recursively=True,
        )
        module_pass.rewrite_module(op)

        if self.num_chunks > 1:
            BackpropagateStencilShapes().apply(ctx, op)



from __future__ import annotations

import itertools
from collections.abc import Iterable, Sequence
from typing import cast

from xdsl.dialects import builtin
from xdsl.dialects.builtin import (
    AnyIntegerAttr,
    ArrayAttr,
    IntegerAttr,
    IntegerType,
    NoneAttr,
    StringAttr,
)
from xdsl.dialects.csl import ParameterDef, csl
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Dialect,
    Operation,
    ParametrizedAttribute,
    Region,
    SSAValue,
)
from xdsl.irdl import (
    IRDLOperation,
    irdl_attr_definition,
    irdl_op_definition,
    opt_prop_def,
    prop_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.traits import HasParent, IsTerminator, Pure
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa

i16 = builtin.IntegerType(16)


@irdl_attr_definition
class ParamAttribute(ParametrizedAttribute):
    """
    Represents a module parameter that needs to have a type, and may have a value.
    """

    name = "csl_wrapper.param"

    key: ParameterDef[StringAttr]
    value: ParameterDef[IntegerAttr[IntegerType] | NoneAttr]
    type: ParameterDef[IntegerType]

    def print_parameters(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string_literal(self.key.data)
            if not isinstance(self.value, NoneAttr):
                printer.print(" default=")
                printer.print_attribute(self.value)
            else:
                printer.print(" : ")
                printer.print_attribute(self.type)

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> Sequence[Attribute]:
        with parser.in_angle_brackets():
            key = StringAttr(parser.parse_str_literal())
            if parser.parse_optional_keyword("default"):
                parser.parse_punctuation("=")
                val = parser.parse_attribute()
                assert isa(val, AnyIntegerAttr)
                assert isinstance(val.type, IntegerType)
                type = val.type
            else:
                parser.parse_punctuation(":")
                val = NoneAttr()
                type = parser.parse_type()
        return key, val, type

    def verify(self) -> None:
        super().verify()
        if not isinstance(self.value, NoneAttr):
            if self.value.type != self.type:
                raise VerifyException(
                    f"Value expected to be of type {self.type}, found {self.value.type}"
                )


@irdl_op_definition
class ImportOp(IRDLOperation):
    """
    Lightweight wrapper around `csl.import_module` that allows specifying field names directly
    and removes the need for handling structs or setting up struct operands
    """

    name = "csl_wrapper.import"

    ops = var_operand_def()
    module = prop_def(StringAttr)
    fields = prop_def(ArrayAttr[StringAttr])
    result = result_def(csl.ImportedModuleType)

    def __init__(
        self, module: str, field_name_mapping: dict[str, Operation | SSAValue]
    ):
        ops: list[Operation | SSAValue] = []
        fields: list[StringAttr] = []
        for field, op in field_name_mapping.items():
            ops.append(op)
            fields.append(StringAttr(field))

        super().__init__(
            operands=[ops],
            properties={
                "module": StringAttr(module),
                "fields": ArrayAttr(fields),
            },
            result_types=[csl.ImportedModuleType()],
        )

    def verify_(self) -> None:
        if len(self.fields) != len(self.ops):
            raise VerifyException("Number of fields does not match number of operands")


@irdl_op_definition
class ModuleOp(IRDLOperation):
    """
    Wrapper class that manages initialisation of layout and program module.

    Specified properties will be lowered to module params (`csl.param`). As such, all properties
    are passed as BlockArgs to the `layout_module` and `program_module` region. The order in which
    properties are specified is important and must match the order of the block args.

    Additionally, any value yielded by the `layout_module` is passed as a block arg (lowered to `csl.param`)
    to the program module. Order is important here as well. The layout module's `yield` op should be lowered to
    `@set_tile_code`, while the program module's `yield` op should be discarded.

    The layout module has two additional block args `x` and `y` as part of the `@set_tile_code` loop nest.
    Operations using these args need to be lowered to the correct place in the loop nest.

    The program module has the following args (in order):
      * general params:        `width` and `height` followed by everything specified in `params`
      * params from layout:    everything defined by `layout_yield_op.fields`
      * input-output symbols:  any arg from the function lowered into this op, which are exported symbols supporting host-device transfers
    """

    name = "csl_wrapper.module"

    width = prop_def(IntegerAttr)
    height = prop_def(IntegerAttr)
    program_name = opt_prop_def(StringAttr)
    params: ArrayAttr[ParamAttribute] = prop_def(ArrayAttr[ParamAttribute])

    layout_module = region_def("single_block")
    program_module = region_def("single_block")

    def __init__(
        self,
        width: int | IntegerAttr[IntegerType],
        height: int | IntegerAttr[IntegerType],
        params: (
            dict[str, IntegerAttr[IntegerType]] | Sequence[ParamAttribute] | None
        ) = None,
    ):
        if not isinstance(width, IntegerAttr):
            width = IntegerAttr(width, i16)
        if not isinstance(height, IntegerAttr):
            height = IntegerAttr(height, i16)
        if params is None:
            params = []
        elif isinstance(params, dict):
            params = [
                ParamAttribute([StringAttr(name), val, val.type])
                for name, val in params.items()
            ]
        params_attr = ArrayAttr(params)

        super().__init__(
            properties={
                "width": width,
                "height": height,
                "params": params_attr,
            },
            regions=[
                Region(
                    Block(
                        arg_types=itertools.chain(
                            [i16] * 4,
                            (param.type for param in params),
                        )
                    )
                ),
                Region(
                    Block(
                        arg_types=itertools.chain(
                            [i16] * 2,
                            (param.type for param in params),
                        )
                    )
                ),
            ],
        )

    def update_program_block_args(
        self,
        yield_args: Iterable[tuple[str, SSAValue]] | None = None,
        exported_symbols: Iterable[tuple[str | None, Attribute]] | None = None,
    ):
        """
        Update `program_module` BlockArguments by adding
        1. yield op fields (pass None to enable automated retrieval, pass empty list to add no yield op fields)
        2. additional exported symbols
        """
        assert (
            len(self.program_module.block.args)
            == len(self.layout_module.block.args) - 2
            # minus two as layout_module has additional x and y args
        ), "program_module block args should only contain args from properties when calling this function"

        if yield_args is None:
            yield_args = self.layout_yield_op.items()

        for name, op in yield_args:
            arg = self.program_module.block.insert_arg(
                op.type, len(self.program_module.block.args)
            )
            arg.name_hint = name

        if exported_symbols is not None:
            for nam, typ in exported_symbols:
                arg = self.program_module.block.insert_arg(
                    typ, len(self.program_module.block.args)
                )
                if nam is not None:
                    arg.name_hint = nam

    def verify_(self):
        # verify that names are unique
        names: set[str] = {"x", "y", "width", "height"}
        for param in self.params.data:
            if param.key.data in names:
                raise VerifyException(f"Duplicate name in parameters: {param.key.data}")
            names.add(param.key.data)

        # verify that x, y, width, height are i16
        if not all(arg.type == i16 for arg in self.layout_module.block.args[:4]):
            raise VerifyException(
                "The first four arguments of the layout block (x, y, width, height) must be of type i16"
            )

        # verify that block args are of the right type for the provided params
        for arg, param in zip(
            [a.type for a in self.layout_module.block.args[4:]],
            self.params,
            strict=True,
        ):
            if arg != param.type:
                raise ValueError(
                    f"Layout module block arg types do not match for arg {param.key} expected: {param.type} but got: "
                    f"{arg}. Block arg types must correspond to prop types (in order)"
                )

        # verify that the first two program block args (width, height) are correctly typed
        if not all(arg.type == i16 for arg in self.program_module.block.args[:2]):
            raise VerifyException(
                "The first two arguments of the program block (width, height) must be of type i16"
            )

        # verify that params and yielded arguments are typed correctly
        # these may be followed by input-output symbols which we cannot verify, therefore setting `strict=False`
        for got, (name, exp) in zip(
            [a.type for a in self.program_module.block.args[2:]],
            itertools.chain(
                (
                    (param.key.data, cast(Attribute, param.type))
                    for param in self.params
                ),
                ((key, val.type) for key, val in self.layout_yield_op.items()),
            ),
            strict=False,
        ):
            if exp != got:
                raise VerifyException(
                    f"Program module block arg types do not match for arg {name} expected: {exp} but got: {got}. "
                    f"Block arg types must correspond to prop types and layout yield result types (in order)"
                )

    def get_layout_param(self, name: str) -> BlockArgument:
        """
        Retrieve layout block arg for name that is x, y, or one of the properties
        """
        # check static params:
        if name in ("x", "y", "width", "height"):
            return self.layout_module.block.args[
                ("x", "y", "width", "height").index(name)
            ]
        # check module params
        for i, param in enumerate(self.params):
            if param.key.data == name:
                return self.layout_module.block.args[4 + i]
        # not found = value error
        raise ValueError(f"{name} does not refer to a block arg of this layout_module")

    def get_program_param(self, name: str) -> BlockArgument:
        """Retrieve program block arg for name that is one of the properties or a param set up by layout yield"""
        # check static params
        if name in ("width", "height"):
            return self.program_module.block.args[("width", "height").index(name)]
        # check module params
        for i, param in enumerate(self.params):
            if param.key.data == name:
                return self.program_module.block.args[2 + i]
        # check yielded params:
        for i, (key, _) in enumerate(self.layout_yield_op.items()):
            if key == name:
                return self.program_module.block.args[2 + len(self.params) + i]
        # not found = value error
        raise ValueError(f"{name} does not refer to a block arg of this program_module")

    @property
    def layout_yield_op(self) -> YieldOp:
        """
        Get the yield op from the layout module. Used in various places.
        """
        return cast(YieldOp, self.layout_module.block.last_op)

    @property
    def exported_symbols(self) -> Sequence[BlockArgument]:
        """
        Get the exported symbols.
        """
        return self.program_module.block.args[
            2 + len(self.params) + len(self.layout_yield_op.fields) :
        ]


@irdl_op_definition
class YieldOp(IRDLOperation):
    """
    Layout module `yield` ops should be lowered to `@set_tile_code` and must specify `fields`.
    Program module `yield` ops have no particular meaning and specify `fields` here is permitted but undefined.
    """

    name = "csl_wrapper.yield"

    values = var_operand_def(Attribute)
    fields = prop_def(ArrayAttr[StringAttr])

    traits = traits_def(
        lambda: frozenset([IsTerminator(), HasParent(ModuleOp), Pure()])
    )

    def __init__(
        self,
        operands: Sequence[SSAValue | Operation],
        args: Sequence[str] | ArrayAttr[StringAttr],
    ):
        if not isinstance(args, ArrayAttr):
            args = ArrayAttr(StringAttr(arg) for arg in args)
        super().__init__(
            operands=[operands],
            properties={
                "fields": args,
            },
        )

    @staticmethod
    def from_field_name_mapping(field_name_mapping: dict[str, Operation | SSAValue]):
        operands: list[Operation | SSAValue] = []
        attributes: list[StringAttr] = []
        for attr, op in field_name_mapping.items():
            attributes.append(StringAttr(attr))
            operands.append(op)
        return YieldOp(operands, ArrayAttr[StringAttr](attributes))

    def verify_(self) -> None:
        if len(self.fields) != len(self.operands):
            raise VerifyException("Number of fields must match the number of operands")

    def items(self) -> Iterable[tuple[str, SSAValue]]:
        assert self.fields is not None
        return zip((elm.data for elm in self.fields.data), self.operands, strict=True)


CSL_WRAPPER = Dialect(
    "csl_wrapper",
    [
        ImportOp,
        ModuleOp,
        YieldOp,
    ],
    [
        ParamAttribute,
    ],
)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from inspect import isclass
from typing import Generic, TypeAlias, TypeVar

from xdsl.ir import Attribute, AttributeCovT, ParametrizedAttribute
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.runtime_final import is_runtime_final


@dataclass
class ConstraintContext:
    """
    Contains the assignment of constraint variables.
    """

    variables: dict[str, Attribute] = field(default_factory=dict)
    """The assignment of constraint variables."""

    def copy(self):
        return ConstraintContext(self.variables.copy())

    def update(self, other: ConstraintContext):
        self.variables.update(other.variables)


_AttributeCovT = TypeVar("_AttributeCovT", bound=Attribute, covariant=True)


@dataclass(frozen=True)
class GenericAttrConstraint(Generic[AttributeCovT], ABC):
    """Constrain an attribute to a certain value."""

    @abstractmethod
    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the attribute satisfies the constraint,
        or raise an exception otherwise.
        """
        ...

    def get_resolved_variables(self) -> set[str]:
        """
        Get the set of type variables that are always resolved when verifying
        the constraint.
        """
        return set()

    def can_infer(self, constraint_names: set[str]) -> bool:
        """
        Check if there is enough information to infer the attribute given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(self, constraint_context: ConstraintContext) -> Attribute:
        """
        Infer the attribute given the constraint variables that are already set.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError("Cannot infer attribute from constraint")

    def get_unique_base(self) -> type[Attribute] | None:
        """Get the unique base type that can satisfy the constraint, if any."""
        return None

    def __or__(
        self, value: GenericAttrConstraint[_AttributeCovT], /
    ) -> AnyOf[AttributeCovT | _AttributeCovT]:
        return AnyOf((self, value))


AttrConstraint: TypeAlias = GenericAttrConstraint[Attribute]


@dataclass(frozen=True)
class VarConstraint(GenericAttrConstraint[AttributeCovT]):
    """
    Constraint variable. If the variable is already set, this will constrain
    the attribute to be equal to the variable. Otherwise, it will first check that the
    variable satisfies the variable constraint, then set the variable with the
    attribute.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: GenericAttrConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if self.name in constraint_context.variables:
            if attr != constraint_context.variables[self.name]:
                raise VerifyException(
                    f"attribute {constraint_context.variables[self.name]} expected from variable "
                    f"'{self.name}', but got {attr}"
                )
        else:
            self.constraint.verify(attr, constraint_context)
            constraint_context.variables[self.name] = attr

    def get_resolved_variables(self) -> set[str]:
        return {self.name, *self.constraint.get_resolved_variables()}

    def can_infer(self, constraint_names: set[str]) -> bool:
        return self.name in constraint_names

    def infer(self, constraint_context: ConstraintContext) -> Attribute:
        constraint_context = constraint_context or ConstraintContext()
        if self.name not in constraint_context.variables:
            raise ValueError(f"Cannot infer attribute from constraint {self}")
        return constraint_context.variables[self.name]

    def get_unique_base(self) -> type[Attribute] | None:
        return self.constraint.get_unique_base()


@dataclass(frozen=True, init=True)
class ConstraintVar:
    """
    Annotation used in PyRDL to define a constraint variable.
    For instance, the following code defines a constraint variable T,
    that can then be used in PyRDL:
    ```python
    T = Annotated[PyRDLConstraint, ConstraintVar("T")]
    ```
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""


@dataclass(frozen=True)
class EqAttrConstraint(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Constrain an attribute to be equal to another attribute."""

    attr: AttributeCovT
    """The attribute we want to check equality with."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if attr != self.attr:
            raise VerifyException(f"Expected attribute {self.attr} but got {attr}")

    def can_infer(self, constraint_names: set[str]) -> bool:
        return True

    def infer(self, constraint_context: ConstraintContext) -> Attribute:
        return self.attr

    def get_unique_base(self) -> type[Attribute] | None:
        return type(self.attr)


@dataclass(frozen=True)
class BaseAttr(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Constrain an attribute to be of a given base type."""

    attr: type[AttributeCovT]
    """The expected attribute base type."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.attr.name}"
            )

    def get_unique_base(self) -> type[Attribute] | None:
        if is_runtime_final(self.attr):
            return self.attr
        return None


def attr_constr_coercion(
    attr: AttributeCovT | type[AttributeCovT] | GenericAttrConstraint[AttributeCovT],
) -> GenericAttrConstraint[AttributeCovT]:
    """
    Attributes are coerced into EqAttrConstraints,
    and Attribute types are coerced into BaseAttr.
    """
    if isinstance(attr, GenericAttrConstraint):
        return attr
    if isinstance(attr, Attribute):
        return EqAttrConstraint(attr)
    if isclass(attr):
        return BaseAttr(attr)
    assert False


@dataclass(frozen=True)
class AnyAttr(GenericAttrConstraint[Attribute]):
    """Constraint that is verified by all attributes."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        pass


@dataclass(frozen=True, init=False)
class AnyOf(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies one of the given constraints."""

    attr_constrs: tuple[GenericAttrConstraint[Attribute], ...]
    """The list of constraints that are checked."""

    def __init__(
        self,
        attr_constrs: Sequence[
            Attribute | type[Attribute] | GenericAttrConstraint[Attribute]
        ],
    ):
        constrs: tuple[GenericAttrConstraint[Attribute], ...] = tuple(
            attr_constr_coercion(constr) for constr in attr_constrs
        )
        object.__setattr__(
            self,
            "attr_constrs",
            constrs,
        )

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext | None = None,
    ) -> None:
        constraint_context = constraint_context or ConstraintContext()
        for attr_constr in self.attr_constrs:
            # Copy the constraint to ensure that if the constraint fails, the
            # constraint context is not modified.
            constraint_context_copy = constraint_context.copy()
            try:
                attr_constr.verify(attr, constraint_context_copy)
                # If the constraint succeeds, we update back the constraint variables
                constraint_context.update(constraint_context_copy)
                return
            except VerifyException:
                pass
        raise VerifyException(f"Unexpected attribute {attr}")

    def __or__(
        self, value: GenericAttrConstraint[_AttributeCovT], /
    ) -> AnyOf[AttributeCovT | _AttributeCovT]:
        return AnyOf((*self.attr_constrs, value))

    def get_resolved_variables(self) -> set[str]:
        if len(self.attr_constrs) == 0:
            return set()
        return set[str].intersection(
            *(constr.get_resolved_variables() for constr in self.attr_constrs)
        )

    def get_unique_base(self) -> type[Attribute] | None:
        bases = [constr.get_unique_base() for constr in self.attr_constrs]
        if None in bases:
            return None
        if len(set(bases)) == 1:
            return bases[0]
        return None


@dataclass(frozen=True)
class AllOf(GenericAttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies all the given constraints."""

    attr_constrs: tuple[GenericAttrConstraint[AttributeCovT], ...]
    """The list of constraints that are checked."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        exc_bucket: list[VerifyException] = []

        for attr_constr in self.attr_constrs:
            try:
                attr_constr.verify(attr, constraint_context)
            except VerifyException as e:
                exc_bucket.append(e)

        if len(exc_bucket):
            if len(exc_bucket) == 1:
                raise VerifyException(str(exc_bucket[0])) from exc_bucket[0]
            exc_msg = "The following constraints were not satisfied:\n"
            exc_msg += "\n".join([str(e) for e in exc_bucket])
            raise VerifyException(exc_msg)

    def get_resolved_variables(self) -> set[str]:
        if len(self.attr_constrs) == 0:
            return set()
        return set[str].union(
            *[constr.get_resolved_variables() for constr in self.attr_constrs]
        )

    def can_infer(self, constraint_names: set[str]) -> bool:
        return any(constr.can_infer(constraint_names) for constr in self.attr_constrs)

    def infer(self, constraint_context: ConstraintContext | None = None) -> Attribute:
        constraint_context = constraint_context or ConstraintContext()
        for constr in self.attr_constrs:
            if constr.can_infer(set(constraint_context.variables.keys())):
                return constr.infer(constraint_context)
        raise ValueError("Cannot infer attribute from constraint")

    def get_unique_base(self) -> type[Attribute] | None:
        # This could be improved if we keep track of all the possible base types for
        # each constraint.
        for constr in self.attr_constrs:
            base = constr.get_unique_base()
            if base is not None:
                return base
        return None


ParametrizedAttributeCovT = TypeVar(
    "ParametrizedAttributeCovT", bound=ParametrizedAttribute, covariant=True
)


@dataclass(frozen=True, init=False)
class ParamAttrConstraint(
    Generic[ParametrizedAttributeCovT], GenericAttrConstraint[ParametrizedAttributeCovT]
):
    """
    Constrain an attribute to be of a given type,
    and also constrain its parameters with additional constraints.
    """

    base_attr: type[ParametrizedAttributeCovT]
    """The base attribute type."""

    param_constrs: tuple[GenericAttrConstraint[Attribute], ...]
    """The attribute parameter constraints"""

    def __init__(
        self,
        base_attr: type[ParametrizedAttributeCovT],
        param_constrs: Sequence[
            (Attribute | type[Attribute] | GenericAttrConstraint[Attribute])
        ],
    ):
        constrs = tuple(attr_constr_coercion(constr) for constr in param_constrs)
        object.__setattr__(self, "base_attr", base_attr)
        object.__setattr__(self, "param_constrs", constrs)

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.base_attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.base_attr.name}"
            )
        if len(self.param_constrs) != len(attr.parameters):
            raise VerifyException(
                f"{len(self.param_constrs)} parameters expected, "
                f"but got {len(attr.parameters)}"
            )
        for idx, param_constr in enumerate(self.param_constrs):
            param_constr.verify(attr.parameters[idx], constraint_context)

    def get_resolved_variables(self) -> set[str]:
        if not self.param_constrs:
            return set()
        return {
            var
            for constr in self.param_constrs
            for var in constr.get_resolved_variables()
        }

    def get_unique_base(self) -> type[Attribute] | None:
        if is_runtime_final(self.base_attr):
            return self.base_attr
        return None


@dataclass(frozen=True, init=False)
class MessageConstraint(GenericAttrConstraint[AttributeCovT]):
    """
    Attach a message to a constraint, to provide more context when the constraint
    is not satisfied.
    """

    constr: GenericAttrConstraint[AttributeCovT]
    message: str

    def __init__(
        self,
        constr: (
            GenericAttrConstraint[AttributeCovT] | AttributeCovT | type[AttributeCovT]
        ),
        message: str,
    ):
        object.__setattr__(self, "constr", attr_constr_coercion(constr))
        object.__setattr__(self, "message", message)

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        try:
            return self.constr.verify(attr, constraint_context)
        except VerifyException as e:
            raise VerifyException(
                f"{self.message}\nUnderlying verification failure: {e.args[0]}",
                *e.args[1:],
            )

    def get_resolved_variables(self) -> set[str]:
        return self.constr.get_resolved_variables()

    def get_unique_base(self) -> type[Attribute] | None:
        return self.constr.get_unique_base()

    def can_infer(self, constraint_names: set[str]) -> bool:
        return self.constr.can_infer(constraint_names)

    def infer(self, constraint_context: ConstraintContext) -> Attribute:
        return self.constr.infer(constraint_context)


class RangeConstraint(Generic[AttributeCovT], ABC):

    @abstractmethod
    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the range satisfies the constraint, or raise an exception otherwise.
        The range can contain Nones, which represent an attribute not to be checked.
        """
        ...

    def get_resolved_variables(self) -> set[str]:
        """
        Get the set of type variables that are always resolved when verifying
        the constraint.
        """
        return set()

    def can_infer(self, constraint_names: set[str]) -> bool:
        """
        Check if there is enough information to infer the range given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(
        self, length: int, constraint_context: ConstraintContext
    ) -> list[Attribute]:
        """
        Infer the range given the constraint variables that are already set.

        Raises an exception if the range cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError("Cannot infer range from constraint")


@dataclass
class RangeOf(RangeConstraint[AttributeCovT]):
    """
    Constrain each element in a range to satisfy a given constraint.
    """

    constr: GenericAttrConstraint[AttributeCovT]

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        for a in attrs:
            self.constr.verify(a, constraint_context)

    def get_resolved_variables(self) -> set[str]:
        return self.constr.get_resolved_variables()

    def can_infer(self, constraint_names: set[str]) -> bool:
        return self.constr.can_infer(constraint_names)

    def infer(
        self, length: int, constraint_context: ConstraintContext
    ) -> list[Attribute]:
        return [self.constr.infer(constraint_context)] * length


@dataclass
class SingleOf(RangeConstraint[AttributeCovT]):
    """
    Constrain a range to only contain a single element, which should satisfy a given constraint.
    """

    constr: GenericAttrConstraint[AttributeCovT]

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        if len(attrs) != 1:
            raise VerifyException(f"Expected a single attribute, got {len(attrs)}")
        self.constr.verify(attrs[0], constraint_context)

    def get_resolved_variables(self) -> set[str]:
        return self.constr.get_resolved_variables()

    def can_infer(self, constraint_names: set[str]) -> bool:
        return self.constr.can_infer(constraint_names)

    def infer(
        self, length: int, constraint_context: ConstraintContext
    ) -> list[Attribute]:
        return [self.constr.infer(constraint_context)]


def range_constr_coercion(
    attr: (
        AttributeCovT
        | type[AttributeCovT]
        | GenericAttrConstraint[AttributeCovT]
        | RangeConstraint[AttributeCovT]
    ),
) -> RangeConstraint[AttributeCovT]:
    if isinstance(attr, RangeConstraint):
        return attr
    return RangeOf(attr_constr_coercion(attr))


def single_range_constr_coercion(
    attr: AttributeCovT | type[AttributeCovT] | GenericAttrConstraint[AttributeCovT],
) -> RangeConstraint[AttributeCovT]:
    return SingleOf(attr_constr_coercion(attr))



from xdsl.context import MLContext
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import PatternRewriter, PatternRewriteWalker, RewritePattern
from xdsl.traits import (
    IsTerminator,
    MemoryEffectKind,
    SymbolOpInterface,
    get_effects,
)


def is_trivially_dead(op: Operation):
    """
    Returns if the operation has no observable effect.
    """
    return (
        all(not result.uses for result in op.results)
        and (not op.get_trait(IsTerminator))
        and (not op.get_trait(SymbolOpInterface))
        and result_only_effects(op)
    )


def result_only_effects(rootOp: Operation) -> bool:
    """
    Returns if we can ensure the operation would have no observable effect beyond its
    returned values.

    cf MLIR's WouldOpBeTriviallyDead:
    https://mlir.llvm.org/doxygen/namespacemlir.html#a655db45ed8c23d04d5ed5ee0abe041ad
    """
    effects = get_effects(rootOp)
    return effects is not None and all(e == MemoryEffectKind.READ for e in effects)


class RemoveUnusedOperations(RewritePattern):
    """
    Removes operations annotated with the `Pure` trait, where results have no uses.
    """

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
        if is_trivially_dead(op):
            rewriter.erase_op(op)


def dce(op: ModuleOp):
    """
    Removes operations annotated with the `Pure` trait, where results have no uses.
    Modifies input module in-place.
    """
    walker = PatternRewriteWalker(
        RemoveUnusedOperations(), apply_recursively=True, walk_reverse=True
    )
    walker.rewrite_module(op)


class DeadCodeElimination(ModulePass):
    name = "dce"

    def apply(self, ctx: MLContext, op: ModuleOp) -> None:
        dce(op)



from collections.abc import Sequence
from itertools import chain
from types import ModuleType
from typing import Any

from xdsl.dialects import get_all_dialects
from xdsl.traits import SymbolTable
from xdsl.utils.dialect_stub import DialectStubGenerator

from . import _version

__version__ = _version.get_versions()["version"]

import importlib.abc
import importlib.util
import os
import sys


class IRDLDialectLoader(importlib.abc.Loader):
    """
    Custom module loader for IRDL files.

    When loading an irdl file as a module:
    - parse the dialect
    - generate its PyRDL implementation
    - generate type stubs for its implmentation and write to disk
    - populate the Python module
    """

    def __init__(self, module_name: str, path: str):
        self.module_name = module_name
        self.path = path

    def exec_module(self, module: ModuleType):
        from xdsl.context import MLContext
        from xdsl.dialects.irdl import DialectOp
        from xdsl.interpreters.irdl import make_dialect
        from xdsl.parser import Parser

        # Open the irdl file
        with open(self.path) as file:
            # Parse it
            ctx = MLContext()
            for dialect_name, dialect_factory in get_all_dialects().items():
                ctx.register_dialect(dialect_name, dialect_factory)
            irdl_module = Parser(ctx, file.read(), self.path).parse_module()

            # Make it a PyRDL Dialect
            filename = os.path.basename(self.path)
            dialect_name, _ = os.path.splitext(filename)
            dialect_op = SymbolTable.lookup_symbol(irdl_module, dialect_name)
            assert isinstance(dialect_op, DialectOp)
            dialect = make_dialect(dialect_op)
            with open(
                f"{os.path.dirname(self.path)}/{dialect_name}.pyi", "w"
            ) as stubfile:
                print(
                    f"""\
""\"
This file is automatically generated by xDSL and not meant to be modified.

It was generated from {self.path}
""\"
""",
                    file=stubfile,
                )
                print(
                    DialectStubGenerator(dialect).generate_dialect_stubs(),
                    file=stubfile,
                )

            for obj in chain(dialect.attributes, dialect.operations):
                setattr(module, obj.__name__, obj)
            setattr(module, dialect.name.capitalize(), dialect)


class IRDLDialectFinder(importlib.abc.MetaPathFinder):
    """
    Custom module finder for IRDL files.

    Look for a <name>.irdl file instead of a <name>.py file.

    """

    def find_spec(self, fullname: str, path: Sequence[str] | None, target: Any = None):
        # Check if module is already loaded and return it if so
        if fullname in sys.modules:
            return sys.modules[fullname].__spec__

        # Look for the file
        filename = fullname.split(".")[-1] + ".irdl"
        if path is None:
            path = [os.getcwd()]
        for entry in path:
            potential_path = os.path.join(entry, filename)
            if os.path.isfile(potential_path):
                # If found, create the right loader and return it
                loader = IRDLDialectLoader(fullname, potential_path)
                return importlib.util.spec_from_file_location(
                    fullname, potential_path, loader=loader
                )

        # Return None if not found to let importlib do its thing.
        return None


# Add the IRDLDialectFinder to the meta path as last resort, i.e, it will look for a
# .irdl implementation if no .py implementation is found.
sys.meta_path.append(IRDLDialectFinder())



import os
import typing
from dataclasses import dataclass, field
from importlib import import_module
from types import ModuleType
from typing import Any

import xdsl.dialects.builtin
import xdsl.ir
import xdsl.irdl
from xdsl.dialects.builtin import ArrayAttr, ArrayOfConstraint
from xdsl.ir import Attribute, Dialect, OpResult, ParametrizedAttribute, Region
from xdsl.irdl import (
    AllOf,
    AnyAttr,
    AnyOf,
    AttrConstraint,
    AttributeDef,
    BaseAttr,
    EqAttrConstraint,
    IRDLOperation,
    OperandDef,
    OptAttributeDef,
    OptOperandDef,
    OptPropertyDef,
    OptRegionDef,
    OptResultDef,
    OptSuccessorDef,
    ParamAttrConstraint,
    PropertyDef,
    RegionDef,
    ResultDef,
    SuccessorDef,
    VarOperandDef,
    VarRegionDef,
    VarResultDef,
    VarSuccessorDef,
)


@dataclass
class DialectStubGenerator:
    """Generate a typing stub file (.pyi) for a dialect."""

    dialect: Dialect
    dependencies: dict[str, set[str]] = field(init=False, default_factory=dict)

    def _import(self, module: ModuleType | str, name: str | type[Any]):
        """
        Internal helper to keep track of dependencies, to later generate clean import
        statements.
        """
        # If passed a type, use its name.
        if isinstance(name, type):
            name = name.__name__
        # If passed a module, use its name.
        if isinstance(module, ModuleType):
            module = module.__name__
        # Do not import from builtins, those are the implicitely available ones.
        if module == "builtins":
            return
        # Do no import from banned nested modules
        if module.startswith("xdsl.ir."):
            module = "xdsl.ir"
        if module.startswith("xdsl.irdl."):
            module = "xdsl.irdl"
        # Create a module dependency or add a new name to the ones from an existing
        # dependency
        if module in self.dependencies:
            self.dependencies[module].add(name)
        else:
            self.dependencies[module] = {name}

    def _generate_constraint_type(self, constraint: AttrConstraint) -> str:
        """
        Return a type hint for the member constrained by a constraint, by it an attribute
        parameter, or an operation attribute/property.
        """
        match constraint:
            case BaseAttr(attr_type):
                if attr_type not in self.dialect.attributes:
                    self._import(attr_type.__module__, attr_type.__name__)
                return attr_type.__name__
            case EqAttrConstraint(attr):
                if type(attr) not in self.dialect.attributes:
                    self._import(type(attr).__module__, type(attr).__name__)
                return type(attr).__name__

            case AnyOf(constraints):
                return " | ".join(
                    self._generate_constraint_type(c) for c in constraints
                )
            case AllOf(constraints):
                self._import(typing, "Annotated")
                return f"Annotated[{', '.join(self._generate_constraint_type(c) for c in reversed(constraints))}]"
            case ArrayOfConstraint(constraint):
                self._import(xdsl.dialects.builtin, ArrayAttr)
                return f"ArrayAttr[{self._generate_constraint_type(constraint)}]"
            case AnyAttr():
                self._import(xdsl.ir, Attribute)
                return "Attribute"
            case ParamAttrConstraint(base_type):
                return base_type.__name__

            case _:
                raise NotImplementedError(
                    f"Unsupported constraint type: {type(constraint)}"
                )

    def _generate_attribute_stub(self, attr: type[ParametrizedAttribute]):
        """
        Generate type stub for an irdl attribute.
        """
        # They all are ParametrizedAttributes.
        self._import(xdsl.ir, ParametrizedAttribute)
        # Get the bases that are not already bases of ParametrizedAttribute.
        bases = set(attr.__mro__[1:]) - set(ParametrizedAttribute.__mro__)
        # Add them as stub dependencies
        for base in bases:
            self._import(base.__module__, base)

        # Also add them to the Attribute class' bases.
        bases = ", ".join(b.__name__ for b in bases)
        if bases:
            bases += ", "
        yield f"class {attr.__name__}({bases}ParametrizedAttribute):"

        # Generate the parameters' stubs, if any
        attr_def = attr.get_irdl_definition()
        for name, param in attr_def.parameters:
            yield f'    {name} : "{self._generate_constraint_type(param)}"'
        # Otherwise, generate a pass for Python's indentation
        if not attr_def.parameters:
            yield "    pass"
        yield ""
        yield ""

    def _generate_operation_stub(self, op: type[IRDLOperation]):
        """
        Generate type stub for an irdl operation.
        """
        # Keep track of whether the operation has any body, to generate a pass if it does not.
        had_body = False

        # They all are IRDLOperations.
        self._import(xdsl.irdl, IRDLOperation)

        # Currently there's nothing that should be a base class or the IRDLOperations.
        # Traits are not supported, and implemented as fields in PyRDL.
        yield f"class {op.__name__}(IRDLOperation):"

        # Generate the constructs' stubs, if any
        op_def = op.get_irdl_definition()
        for name, o in op_def.operands:
            had_body = True
            match o:
                case VarOperandDef(_):
                    self._import(xdsl.irdl, "VarOperand")
                    yield f"    {name} : VarOperand"
                case OptOperandDef(_):
                    self._import(xdsl.irdl, "OptOperand")
                    yield f"    {name} : OptOperand"
                case OperandDef(_):
                    self._import(xdsl.irdl, "Operand")
                    yield f"    {name} : Operand"
        for name, o in op_def.results:
            had_body = True
            match o:
                case VarResultDef():
                    self._import(xdsl.irdl, "VarOpResult")
                    yield f"    {name} : VarOpResult"
                case OptResultDef():
                    self._import(xdsl.irdl, "OptOpResult")
                    yield f"    {name} : OptOpResult"
                case ResultDef():
                    self._import(xdsl.ir, OpResult)
                    yield f"    {name} : OpResult"
        for name, o in op_def.attributes.items():
            had_body = True
            match o:
                case OptAttributeDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)} | None"
                case AttributeDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)}"
        for name, o in op_def.properties.items():
            had_body = True
            match o:
                case OptPropertyDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)} | None"
                case PropertyDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)}"

        for name, r in op_def.regions:
            had_body = True
            match r:
                case OptRegionDef():
                    self._import(xdsl.irdl, "OptRegion")
                    yield f"    {name} : OptRegion"
                case VarRegionDef():
                    self._import(xdsl.irdl, "VarRegion")
                    yield f"    {name} : VarRegion"
                case RegionDef():
                    self._import(xdsl.ir, Region)
                    yield f"    {name} : Region"

        for name, r in op_def.successors:
            had_body = True
            match r:
                case OptSuccessorDef():
                    self._import(xdsl.irdl, "OptSuccessor")
                    yield f"    {name} : OptSuccessor"
                case VarSuccessorDef():
                    self._import(xdsl.irdl, "VarSuccessor")
                    yield f"    {name} : VarSuccessor"
                case SuccessorDef():
                    self._import(xdsl.irdl, "Successor")
                    yield f"    {name} : Successor"
        # Generate a pass if the operation had no body.
        if not had_body:
            yield "    pass"
        yield ""
        yield ""

    def _generate_dialect_stubs(self):
        """
        Generate a dialect's stubs.

        Just generate stubs for all attributes and operations in the dialect.
        """
        for attr in self.dialect.attributes:
            if issubclass(attr, ParametrizedAttribute):
                for l in self._generate_attribute_stub(attr):
                    yield l

        for op in self.dialect.operations:
            if issubclass(op, IRDLOperation):
                for l in self._generate_operation_stub(op):
                    yield l

    def _generate_imports(self):
        """
        Generate import statements for all the dependencies of the stub.
        """
        # sort modules alphabetically for deterministic and clean output.
        items = list(self.dependencies.items())
        items.sort()

        for module, names in items:
            # If only one name is imported from a module, make a one-liner import.
            if len(names) == 1:
                name = names.pop()
                yield f"from {module} import {name}"
            # Otherwise, import all names in a multi-line import, sorted again for
            # a deterministic and clean output.
            else:
                names = list(names)
                names.sort()
                yield f"from {module} import ("
                for o in names:
                    yield f"    {o},"
                yield ")"

    def generate_dialect_stubs(self):
        """
        The main function, generate stubs for the passed dialect and return as a string.

        NB: probably not optimal perf-wise, but I don't foresee this as a bottleneck.
        """
        self._import(xdsl.ir, Dialect)
        dialect_body = "\n".join(self._generate_dialect_stubs())
        imports = "\n".join(self._generate_imports())
        if imports:
            imports += "\n"

        return f"""\
{imports}
{dialect_body}
{self.dialect.name.capitalize()} : Dialect
"""


def make_all_stubs():
    dialects = xdsl.dialects
    directory = "/".join(dialects.__path__)
    for file in os.listdir(directory):
        name, ext = os.path.splitext(file)
        if ext == ".irdl":
            import_module(f"{directory}/{name}")


if __name__ == "__main__":
    make_all_stubs()



from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xdsl.ir import Attribute, Dialect, Operation


@dataclass
class MLContext:
    """Contains structures for operations/attributes registration."""

    allow_unregistered: bool = field(default=False)

    _loaded_dialects: "dict[str, Dialect]" = field(default_factory=dict)
    _loaded_ops: "dict[str, type[Operation]]" = field(default_factory=dict)
    _loaded_attrs: "dict[str, type[Attribute]]" = field(default_factory=dict)
    _registered_dialects: "dict[str, Callable[[], Dialect]]" = field(
        default_factory=dict
    )
    """
    A dictionary of all registered dialects that are not yet loaded. This is used to
    only load the respective Python files when the dialect is actually used.
    """

    def clone(self) -> "MLContext":
        return MLContext(
            self.allow_unregistered,
            self._loaded_dialects.copy(),
            self._loaded_ops.copy(),
            self._loaded_attrs.copy(),
            self._registered_dialects.copy(),
        )

    @property
    def loaded_ops(self) -> "Iterable[type[Operation]]":
        """
        Returns all the loaded operations. Not valid across mutations of this object.
        """
        return self._loaded_ops.values()

    @property
    def loaded_attrs(self) -> "Iterable[type[Attribute]]":
        """
        Returns all the loaded attributes. Not valid across mutations of this object.
        """
        return self._loaded_attrs.values()

    @property
    def loaded_dialects(self) -> "Iterable[Dialect]":
        """
        Returns all the loaded attributes. Not valid across mutations of this object.
        """
        return self._loaded_dialects.values()

    @property
    def registered_dialect_names(self) -> Iterable[str]:
        """
        Returns the names of all registered dialects. Not valid across mutations of this object.
        """
        return self._registered_dialects.keys()

    def register_dialect(
        self, name: str, dialect_factory: "Callable[[], Dialect]"
    ) -> None:
        """
        Register a dialect without loading it. The dialect is only loaded in the context
        when an operation or attribute of that dialect is parsed, or when explicitely
        requested with `load_registered_dialect`.
        """
        if name in self._registered_dialects:
            raise ValueError(f"'{name}' dialect is already registered")
        self._registered_dialects[name] = dialect_factory

    def load_registered_dialect(self, name: str) -> None:
        """Load a dialect that is already registered in the context."""
        if name not in self._registered_dialects:
            raise ValueError(f"'{name}' dialect is not registered")
        dialect = self._registered_dialects[name]()
        self._loaded_dialects[dialect.name] = dialect

        for op in dialect.operations:
            self.load_op(op)

        for attr in dialect.attributes:
            self.load_attr(attr)

    def load_dialect(self, dialect: "Dialect"):
        """
        Load a dialect. Operation and Attribute names should be unique.
        If the dialect is already registered in the context, use
        `load_registered_dialect` instead.
        """
        if dialect.name in self._registered_dialects:
            raise ValueError(
                f"'{dialect.name}' dialect is already registered, use 'load_registered_dialect' instead"
            )
        self.register_dialect(dialect.name, lambda: dialect)
        self.load_registered_dialect(dialect.name)

    def load_op(self, op: "type[Operation]") -> None:
        """Load an operation definition. Operation names should be unique."""
        if op.name in self._loaded_ops:
            raise Exception(f"Operation {op.name} has already been loaded")
        self._loaded_ops[op.name] = op

    def load_attr(self, attr: "type[Attribute]") -> None:
        """Load an attribute definition. Attribute names should be unique."""
        if attr.name in self._loaded_attrs:
            raise Exception(f"Attribute {attr.name} has already been loaded")
        self._loaded_attrs[attr.name] = attr

    def get_optional_op(self, name: str) -> "type[Operation] | None":
        """
        Get an operation class from its name if it exists.
        If the operation is not registered, return None unless unregistered operations
        are allowed in the context, in which case return an UnregisteredOp.
        """
        # If the operation is already loaded, returns it.
        if name in self._loaded_ops:
            return self._loaded_ops[name]

        # Otherwise, check if the operation dialect is registered.
        if "." in name:
            dialect_name, _ = name.split(".", 1)
            if (
                dialect_name in self._registered_dialects
                and dialect_name not in self._loaded_dialects
            ):
                self.load_registered_dialect(dialect_name)
                return self.get_optional_op(name)

        # If the dialect is unregistered, but the context allows unregistered
        # operations, return an UnregisteredOp.
        if self.allow_unregistered:
            from xdsl.dialects.builtin import UnregisteredOp

            op_type = UnregisteredOp.with_name(name)
            self._loaded_ops[name] = op_type
            return op_type
        return None

    def get_op(self, name: str) -> "type[Operation]":
        """
        Get an operation class from its name.
        If the operation is not registered, raise an exception unless unregistered
        operations are allowed in the context, in which case return an UnregisteredOp.
        """
        if op_type := self.get_optional_op(name):
            return op_type
        raise Exception(f"Operation {name} is not registered")

    def get_optional_attr(
        self,
        name: str,
        create_unregistered_as_type: bool = False,
    ) -> "type[Attribute] | None":
        """
        Get an attribute class from its name if it exists.
        If the attribute is not registered, return None unless unregistered attributes
        are allowed in the context, in which case return an UnregisteredAttr.
        Since UnregisteredAttr may be a type (for MLIR compatibility), an
        additional flag is required to create an UnregisterAttr that is
        also a type.
        """
        # If the attribute is already loaded, returns it.
        if name in self._loaded_attrs:
            return self._loaded_attrs[name]

        # Otherwise, check if the attribute dialect is registered.
        dialect_name, _ = name.split(".", 1)
        if (
            dialect_name in self._registered_dialects
            and dialect_name not in self._loaded_dialects
        ):
            self.load_registered_dialect(dialect_name)
            return self.get_optional_attr(name)

        # If the dialect is unregistered, but the context allows unregistered
        # attributes, return an UnregisteredOp.
        if self.allow_unregistered:
            from xdsl.dialects.builtin import UnregisteredAttr

            attr_type = UnregisteredAttr.with_name_and_type(
                name, create_unregistered_as_type
            )
            self._loaded_attrs[name] = attr_type
            return attr_type

        return None

    def get_attr(
        self,
        name: str,
        create_unregistered_as_type: bool = False,
    ) -> "type[Attribute]":
        """
        Get an attribute class from its name.
        If the attribute is not registered, raise an exception unless unregistered
        attributes are allowed in the context, in which case return an UnregisteredAttr.
        Since UnregisteredAttr may be a type (for MLIR compatibility), an
        additional flag is required to create an UnregisterAttr that is
        also a type.
        """
        if attr_type := self.get_optional_attr(name, create_unregistered_as_type):
            return attr_type
        raise Exception(f"Attribute {name} is not registered")

    def get_dialect(self, name: str) -> "Dialect":
        if (dialect := self.get_optional_dialect(name)) is None:
            raise Exception(f"Dialect {name} is not registered")
        return dialect

    def get_optional_dialect(self, name: str) -> "Dialect | None":
        if name in self._loaded_dialects:
            return self._loaded_dialects[name]
        return None



import argparse
import os
import sys
from collections.abc import Callable
from typing import IO

from xdsl.context import MLContext
from xdsl.dialects import get_all_dialects
from xdsl.dialects.builtin import ModuleOp
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.utils.exceptions import ParseError
from xdsl.utils.lexer import Span


def get_all_passes() -> dict[str, Callable[[], type[ModulePass]]]:
    """Return the list of all available passes."""

    def get_arith_add_fastmath():
        from xdsl.transforms import arith_add_fastmath

        return arith_add_fastmath.AddArithFastMathFlagsPass

    def get_loop_hoist_memref():
        from xdsl.transforms import loop_hoist_memref

        return loop_hoist_memref.LoopHoistMemrefPass

    def get_canonicalize():
        from xdsl.transforms import canonicalize

        return canonicalize.CanonicalizePass

    def get_canonicalize_dmp():
        from xdsl.transforms import canonicalize_dmp

        return canonicalize_dmp.CanonicalizeDmpPass

    def get_convert_linalg_to_memref_stream():
        from xdsl.transforms import convert_linalg_to_memref_stream

        return convert_linalg_to_memref_stream.ConvertLinalgToMemrefStreamPass

    def get_convert_linalg_to_loops():
        from xdsl.transforms import convert_linalg_to_loops

        return convert_linalg_to_loops.ConvertLinalgToLoopsPass

    def get_convert_ml_program_to_memref():
        from xdsl.transforms import convert_ml_program_to_memref

        return convert_ml_program_to_memref.ConvertMlProgramToMemrefPass

    def get_convert_riscv_scf_for_to_frep():
        from xdsl.transforms import convert_riscv_scf_for_to_frep

        return convert_riscv_scf_for_to_frep.ConvertRiscvScfForToFrepPass

    def get_convert_scf_to_openmp():
        from xdsl.transforms import convert_scf_to_openmp

        return convert_scf_to_openmp.ConvertScfToOpenMPPass

    def get_convert_snitch_stream_to_snitch():
        from xdsl.backend.riscv.lowering import convert_snitch_stream_to_snitch

        return convert_snitch_stream_to_snitch.ConvertSnitchStreamToSnitch

    def get_constant_fold_interp():
        from xdsl.transforms import constant_fold_interp

        return constant_fold_interp.ConstantFoldInterpPass

    def get_convert_snrt_to_riscv():
        from xdsl.transforms import inline_snrt

        return inline_snrt.ConvertSnrtToRISCV

    def get_convert_stencil_to_ll_mlir():
        from xdsl.transforms.experimental import convert_stencil_to_ll_mlir

        return convert_stencil_to_ll_mlir.ConvertStencilToLLMLIRPass

    def get_convert_riscv_scf_to_riscv_cf():
        from xdsl.backend.riscv.lowering import convert_riscv_scf_to_riscv_cf

        return convert_riscv_scf_to_riscv_cf.ConvertRiscvScfToRiscvCfPass

    def get_cse():
        from xdsl.transforms import common_subexpression_elimination

        return common_subexpression_elimination.CommonSubexpressionElimination

    def get_csl_stencil_to_csl_wrapper():
        from xdsl.transforms import csl_stencil_to_csl_wrapper

        return csl_stencil_to_csl_wrapper.CslStencilToCslWrapperPass

    def get_dce():
        from xdsl.transforms import dead_code_elimination

        return dead_code_elimination.DeadCodeElimination

    def get_desymrefy():
        from xdsl.frontend.passes.desymref import DesymrefyPass

        return DesymrefyPass

    def get_gpu_allocs():
        from xdsl.transforms import gpu_allocs

        return gpu_allocs.MemrefToGPUPass

    def get_gpu_map_parallel_loops():
        from xdsl.transforms import gpu_map_parallel_loops

        return gpu_map_parallel_loops.GpuMapParallelLoopsPass

    def get_distribute_stencil():
        from xdsl.transforms.experimental.dmp import stencil_global_to_local

        return stencil_global_to_local.DistributeStencilPass

    def get_lower_halo_to_mpi():
        from xdsl.transforms.experimental.dmp import stencil_global_to_local

        return stencil_global_to_local.LowerHaloToMPI

    def get_individual_rewrite():
        from xdsl.transforms.individual_rewrite import IndividualRewrite

        return IndividualRewrite

    def get_lower_affine():
        from xdsl.transforms import lower_affine

        return lower_affine.LowerAffinePass

    def get_lower_mpi():
        from xdsl.transforms import lower_mpi

        return lower_mpi.LowerMPIPass

    def get_lower_riscv_func():
        from xdsl.transforms import lower_riscv_func

        return lower_riscv_func.LowerRISCVFunc

    def get_lower_snitch():
        from xdsl.transforms import lower_snitch

        return lower_snitch.LowerSnitchPass

    def get_memref_streamify():
        from xdsl.transforms import memref_streamify

        return memref_streamify.MemrefStreamifyPass

    def get_memref_stream_unnest_out_parameters():
        from xdsl.transforms import memref_stream_unnest_out_parameters

        return memref_stream_unnest_out_parameters.MemrefStreamUnnestOutParametersPass

    def get_memref_stream_fold_fill():
        from xdsl.transforms import memref_stream_fold_fill

        return memref_stream_fold_fill.MemrefStreamFoldFillPass

    def get_memref_stream_generalize_fill():
        from xdsl.transforms import memref_stream_generalize_fill

        return memref_stream_generalize_fill.MemrefStreamGeneralizeFillPass

    def get_memref_stream_infer_fill():
        from xdsl.transforms import memref_stream_infer_fill

        return memref_stream_infer_fill.MemrefStreamInferFillPass

    def get_memref_stream_interleave():
        from xdsl.transforms import memref_stream_interleave

        return memref_stream_interleave.MemrefStreamInterleavePass

    def get_memref_stream_tile_outer_loops():
        from xdsl.transforms import memref_stream_tile_outer_loops

        return memref_stream_tile_outer_loops.MemrefStreamTileOuterLoopsPass

    def get_mlir_opt():
        from xdsl.transforms import mlir_opt

        return mlir_opt.MLIROptPass

    def get_printf_to_llvm():
        from xdsl.transforms import printf_to_llvm

        return printf_to_llvm.PrintfToLLVM

    def get_printf_to_putchar():
        from xdsl.transforms import printf_to_putchar

        return printf_to_putchar.PrintfToPutcharPass

    def get_riscv_register_allocation():
        from xdsl.transforms import riscv_register_allocation

        return riscv_register_allocation.RISCVRegisterAllocation

    def get_scf_for_loop_flatten():
        from xdsl.transforms import scf_for_loop_flatten

        return scf_for_loop_flatten.ScfForLoopFlattenPass

    def get_riscv_scf_loop_range_folding():
        from xdsl.transforms import riscv_scf_loop_range_folding

        return riscv_scf_loop_range_folding.RiscvScfLoopRangeFoldingPass

    def get_snitch_register_allocation():
        from xdsl.transforms import snitch_register_allocation

        return snitch_register_allocation.SnitchRegisterAllocation

    def get_riscv_prologue_epilogue_insertion():
        from xdsl.backend.riscv import prologue_epilogue_insertion

        return prologue_epilogue_insertion.PrologueEpilogueInsertion

    def get_convert_arith_to_riscv():
        from xdsl.backend.riscv.lowering import convert_arith_to_riscv

        return convert_arith_to_riscv.ConvertArithToRiscvPass

    def get_convert_arith_to_riscv_snitch():
        from xdsl.backend.riscv.lowering import convert_arith_to_riscv_snitch

        return convert_arith_to_riscv_snitch.ConvertArithToRiscvSnitchPass

    def get_convert_func_to_riscv_func():
        from xdsl.backend.riscv.lowering import convert_func_to_riscv_func

        return convert_func_to_riscv_func.ConvertFuncToRiscvFuncPass

    def get_convert_memref_to_riscv():
        from xdsl.backend.riscv.lowering import convert_memref_to_riscv

        return convert_memref_to_riscv.ConvertMemrefToRiscvPass

    def get_convert_memref_stream_to_loops():
        from xdsl.transforms import convert_memref_stream_to_loops

        return convert_memref_stream_to_loops.ConvertMemrefStreamToLoopsPass

    def get_convert_onnx_to_linalg():
        from xdsl.transforms import convert_onnx_to_linalg

        return convert_onnx_to_linalg.ConvertOnnxToLinalgPass

    def get_convert_memref_stream_to_snitch_stream():
        from xdsl.transforms import convert_memref_stream_to_snitch_stream

        return (
            convert_memref_stream_to_snitch_stream.ConvertMemrefStreamToSnitchStreamPass
        )

    def get_convert_print_format_to_riscv_debug():
        from xdsl.backend.riscv.lowering import convert_print_format_to_riscv_debug

        return convert_print_format_to_riscv_debug.ConvertPrintFormatToRiscvDebugPass

    def get_convert_qref_to_qssa():
        from xdsl.transforms import convert_qref_to_qssa

        return convert_qref_to_qssa.ConvertQRefToQssa

    def get_convert_qssa_to_qref():
        from xdsl.transforms import convert_qssa_to_qref

        return convert_qssa_to_qref.ConvertQssaToQRef

    def get_scf_parallel_loop_tiling():
        from xdsl.transforms import scf_parallel_loop_tiling

        return scf_parallel_loop_tiling.ScfParallelLoopTilingPass

    def get_convert_scf_to_riscv_scf():
        from xdsl.backend.riscv.lowering import convert_scf_to_riscv_scf

        return convert_scf_to_riscv_scf.ConvertScfToRiscvPass

    def get_function_constant_pinning():
        from xdsl.transforms.experimental.function_constant_pinning import (
            FunctionConstantPinningPass,
        )

        return FunctionConstantPinningPass

    def get_lower_scf_for_to_labels():
        from xdsl.backend.riscv import riscv_scf_to_asm

        return riscv_scf_to_asm.LowerScfForToLabels

    def get_stencil_shape_inference():
        from xdsl.transforms.experimental import stencil_shape_inference

        return stencil_shape_inference.StencilShapeInferencePass

    def get_stencil_storage_materialization():
        from xdsl.transforms.experimental import stencil_storage_materialization

        return stencil_storage_materialization.StencilStorageMaterializationPass

    def get_reconcile_unrealized_casts():
        from xdsl.transforms import reconcile_unrealized_casts

        return reconcile_unrealized_casts.ReconcileUnrealizedCastsPass

    def get_hls_convert_stencil_to_ll_mlir():
        from xdsl.transforms.experimental import hls_convert_stencil_to_ll_mlir

        return hls_convert_stencil_to_ll_mlir.HLSConvertStencilToLLMLIRPass

    def get_lower_hls():
        from xdsl.transforms.experimental import lower_hls

        return lower_hls.LowerHLSPass

    def get_replace_incompatible_fpga():
        from xdsl.transforms.experimental import replace_incompatible_fpga

        return replace_incompatible_fpga.ReplaceIncompatibleFPGA

    def get_stencil_tensorize_z_dimension():
        from xdsl.transforms.experimental import stencil_tensorize_z_dimension

        return stencil_tensorize_z_dimension.StencilTensorizeZDimension

    def get_stencil_to_csl_stencil():
        from xdsl.transforms import stencil_to_csl_stencil

        return stencil_to_csl_stencil.StencilToCslStencilPass

    def get_stencil_unroll():
        from xdsl.transforms import stencil_unroll

        return stencil_unroll.StencilUnrollPass

    def get_test_lower_snitch_stream_to_asm():
        from xdsl.transforms import test_lower_snitch_stream_to_asm

        return test_lower_snitch_stream_to_asm.TestLowerSnitchStreamToAsm

    def get_test_lower_memref_stream_to_snitch_stream():
        from xdsl.transforms import test_lower_memref_stream_to_snitch_stream

        return (
            test_lower_memref_stream_to_snitch_stream.TestLowerMemrefStreamToSnitchStream
        )

    def get_test_optimise_memref_stream():
        from xdsl.transforms import test_optimise_memref_stream

        return test_optimise_memref_stream.TestOptimiseMemrefStream

    return {
        "arith-add-fastmath": get_arith_add_fastmath,
        "loop-hoist-memref": get_loop_hoist_memref,
        "canonicalize-dmp": get_canonicalize_dmp,
        "canonicalize": get_canonicalize,
        "constant-fold-interp": get_constant_fold_interp,
        "convert-arith-to-riscv": get_convert_arith_to_riscv,
        "convert-arith-to-riscv-snitch": get_convert_arith_to_riscv_snitch,
        "convert-func-to-riscv-func": get_convert_func_to_riscv_func,
        "convert-linalg-to-memref-stream": get_convert_linalg_to_memref_stream,
        "convert-linalg-to-loops": get_convert_linalg_to_loops,
        "convert-memref-stream-to-loops": get_convert_memref_stream_to_loops,
        "convert-memref-to-riscv": get_convert_memref_to_riscv,
        "convert-ml-program-to-memref": get_convert_ml_program_to_memref,
        "convert-onnx-to-linalg": get_convert_onnx_to_linalg,
        "convert-memref-stream-to-snitch-stream": get_convert_memref_stream_to_snitch_stream,
        "convert-print-format-to-riscv-debug": get_convert_print_format_to_riscv_debug,
        "convert-qref-to-qssa": get_convert_qref_to_qssa,
        "convert-qssa-to-qref": get_convert_qssa_to_qref,
        "convert-riscv-scf-for-to-frep": get_convert_riscv_scf_for_to_frep,
        "convert-riscv-scf-to-riscv-cf": get_convert_riscv_scf_to_riscv_cf,
        "convert-scf-to-openmp": get_convert_scf_to_openmp,
        "convert-scf-to-riscv-scf": get_convert_scf_to_riscv_scf,
        "convert-snitch-stream-to-snitch": get_convert_snitch_stream_to_snitch,
        "inline-snrt": get_convert_snrt_to_riscv,
        "convert-stencil-to-ll-mlir": get_convert_stencil_to_ll_mlir,
        "cse": get_cse,
        "csl-stencil-to-csl-wrapper": get_csl_stencil_to_csl_wrapper,
        "dce": get_dce,
        "distribute-stencil": get_distribute_stencil,
        "dmp-to-mpi": get_lower_halo_to_mpi,
        "frontend-desymrefy": get_desymrefy,
        "function-constant-pinning": get_function_constant_pinning,
        "memref-to-gpu": get_gpu_allocs,
        "gpu-map-parallel-loops": get_gpu_map_parallel_loops,
        "hls-convert-stencil-to-ll-mlir": get_hls_convert_stencil_to_ll_mlir,
        "apply-individual-rewrite": get_individual_rewrite,
        "lower-affine": get_lower_affine,
        "lower-hls": get_lower_hls,
        "lower-mpi": get_lower_mpi,
        "lower-riscv-func": get_lower_riscv_func,
        "lower-riscv-scf-to-labels": get_lower_scf_for_to_labels,
        "lower-snitch": get_lower_snitch,
        "memref-streamify": get_memref_streamify,
        "memref-stream-unnest-out-parameters": get_memref_stream_unnest_out_parameters,
        "memref-stream-fold-fill": get_memref_stream_fold_fill,
        "memref-stream-generalize-fill": get_memref_stream_generalize_fill,
        "memref-stream-infer-fill": get_memref_stream_infer_fill,
        "memref-stream-interleave": get_memref_stream_interleave,
        "memref-stream-tile-outer-loops": get_memref_stream_tile_outer_loops,
        "mlir-opt": get_mlir_opt,
        "printf-to-llvm": get_printf_to_llvm,
        "printf-to-putchar": get_printf_to_putchar,
        "reconcile-unrealized-casts": get_reconcile_unrealized_casts,
        "replace-incompatible-fpga": get_replace_incompatible_fpga,
        "riscv-allocate-registers": get_riscv_register_allocation,
        "scf-for-loop-flatten": get_scf_for_loop_flatten,
        "riscv-scf-loop-range-folding": get_riscv_scf_loop_range_folding,
        "scf-parallel-loop-tiling": get_scf_parallel_loop_tiling,
        "snitch-allocate-registers": get_snitch_register_allocation,
        "riscv-prologue-epilogue-insertion": get_riscv_prologue_epilogue_insertion,
        "stencil-shape-inference": get_stencil_shape_inference,
        "stencil-storage-materialization": get_stencil_storage_materialization,
        "stencil-tensorize-z-dimension": get_stencil_tensorize_z_dimension,
        "stencil-to-csl-stencil": get_stencil_to_csl_stencil,
        "stencil-unroll": get_stencil_unroll,
        "test-lower-snitch-stream-to-asm": get_test_lower_snitch_stream_to_asm,
        "test-lower-memref-stream-to-snitch-stream": get_test_lower_memref_stream_to_snitch_stream,
        "test-optimise-memref-stream": get_test_optimise_memref_stream,
    }


class CommandLineTool:
    ctx: MLContext
    args: argparse.Namespace
    """
    The argument parsers namespace which holds the parsed commandline
    attributes.
    """

    available_frontends: dict[str, Callable[[IO[str]], ModuleOp]]
    """
    A mapping from file extension to a frontend that can handle this
    file type.
    """

    def register_all_arguments(self, arg_parser: argparse.ArgumentParser):
        arg_parser.add_argument(
            "input_file", type=str, nargs="?", help="path to input file"
        )

        frontends = [name for name in self.available_frontends]
        arg_parser.add_argument(
            "-f",
            "--frontend",
            type=str,
            required=False,
            choices=frontends,
            help="Frontend to be used for the input. If not set, "
            "the xdsl frontend or the one for the file extension "
            "is used.",
        )
        arg_parser.add_argument("--disable-verify", default=False, action="store_true")

        arg_parser.add_argument(
            "--allow-unregistered-dialect",
            default=False,
            action="store_true",
            help="Allow the parsing of unregistered dialects.",
        )

        arg_parser.add_argument(
            "--no-implicit-module",
            default=False,
            action="store_true",
            help="Disable implicit addition of a top-level module op during parsing.",
        )

    def get_input_stream(self) -> tuple[IO[str], str]:
        """
        Get the input stream to parse from, along with the file extension.
        """
        if self.args.input_file is None:
            f = sys.stdin
            file_extension = "mlir"
        else:
            f = open(self.args.input_file)
            _, file_extension = os.path.splitext(self.args.input_file)
            file_extension = file_extension.replace(".", "")
        return f, file_extension

    def get_input_name(self):
        return self.args.input_file or "stdin"

    def register_all_dialects(self):
        """
        Register all dialects that can be used.

        Add other/additional dialects by overloading this function.
        """
        for dialect_name, dialect_factory in get_all_dialects().items():
            self.ctx.register_dialect(dialect_name, dialect_factory)

    def register_all_frontends(self):
        """
        Register all frontends that can be used.

        Add other/additional frontends by overloading this function.
        """

        def parse_mlir(io: IO[str]):
            return Parser(
                self.ctx,
                io.read(),
                self.get_input_name(),
            ).parse_module(not self.args.no_implicit_module)

        self.available_frontends["mlir"] = parse_mlir

    def parse_chunk(
        self, chunk: IO[str], file_extension: str, start_offset: int = 0
    ) -> ModuleOp | None:
        """
        Parse the input file by invoking the parser specified by the `parser`
        argument. If not set, the parser registered for this file extension
        is used.
        """

        try:
            return self.available_frontends[file_extension](chunk)
        except ParseError as e:
            s = e.span
            e.span = Span(s.start, s.end, s.input, start_offset)
            if "parsing_diagnostics" in self.args and self.args.parsing_diagnostics:
                print(e.with_context())
            else:
                raise Exception("Failed to parse:\n" + e.with_context()) from e
        finally:
            chunk.close()



from dataclasses import dataclass

from xdsl.passes import ModulePass, PipelinePass
from xdsl.transforms.canonicalize import CanonicalizePass
from xdsl.transforms.convert_memref_stream_to_loops import (
    ConvertMemrefStreamToLoopsPass,
)
from xdsl.transforms.memref_stream_fold_fill import MemrefStreamFoldFillPass
from xdsl.transforms.memref_stream_generalize_fill import MemrefStreamGeneralizeFillPass
from xdsl.transforms.memref_stream_infer_fill import MemrefStreamInferFillPass
from xdsl.transforms.memref_stream_interleave import MemrefStreamInterleavePass
from xdsl.transforms.memref_stream_tile_outer_loops import (
    MemrefStreamTileOuterLoopsPass,
)
from xdsl.transforms.memref_stream_unnest_out_parameters import (
    MemrefStreamUnnestOutParametersPass,
)
from xdsl.transforms.memref_streamify import MemrefStreamifyPass
from xdsl.transforms.scf_for_loop_flatten import ScfForLoopFlattenPass

TEST_OPTIMISE_MEMREF_STREAM: tuple[ModulePass, ...] = (
    CanonicalizePass(),
    MemrefStreamInferFillPass(),
    MemrefStreamUnnestOutParametersPass(),
    MemrefStreamFoldFillPass(),
    MemrefStreamGeneralizeFillPass(),
    MemrefStreamInterleavePass(),
    MemrefStreamTileOuterLoopsPass(target_rank=4),
    MemrefStreamifyPass(),
    ConvertMemrefStreamToLoopsPass(),
    CanonicalizePass(),
    ScfForLoopFlattenPass(),
)


@dataclass(frozen=True)
class TestOptimiseMemrefStream(PipelinePass):
    """
    A compiler pass used for testing the optimization of memref streams.
    """

    name = "test-optimise-memref-stream"

    passes: tuple[ModulePass, ...] = TEST_OPTIMISE_MEMREF_STREAM



from dataclasses import dataclass

from xdsl.backend.riscv.lowering.convert_arith_to_riscv import ConvertArithToRiscvPass
from xdsl.backend.riscv.lowering.convert_arith_to_riscv_snitch import (
    ConvertArithToRiscvSnitchPass,
)
from xdsl.backend.riscv.lowering.convert_func_to_riscv_func import (
    ConvertFuncToRiscvFuncPass,
)
from xdsl.backend.riscv.lowering.convert_memref_to_riscv import ConvertMemrefToRiscvPass
from xdsl.backend.riscv.lowering.convert_scf_to_riscv_scf import ConvertScfToRiscvPass
from xdsl.passes import ModulePass, PipelinePass
from xdsl.transforms.canonicalize import CanonicalizePass
from xdsl.transforms.convert_memref_stream_to_snitch_stream import (
    ConvertMemrefStreamToSnitchStreamPass,
)
from xdsl.transforms.lower_affine import LowerAffinePass
from xdsl.transforms.reconcile_unrealized_casts import ReconcileUnrealizedCastsPass

TEST_LOWER_MEMREF_STREAM_TO_SNITCH_STREAM: tuple[ModulePass, ...] = (
    CanonicalizePass(),
    ConvertMemrefToRiscvPass(),
    LowerAffinePass(),
    ConvertScfToRiscvPass(),
    ConvertArithToRiscvSnitchPass(),
    ConvertArithToRiscvPass(),
    ConvertFuncToRiscvFuncPass(),
    ConvertMemrefStreamToSnitchStreamPass(),
    ReconcileUnrealizedCastsPass(),
)


@dataclass(frozen=True)
class TestLowerMemrefStreamToSnitchStream(PipelinePass):
    """
    A compiler pass used for testing of the lowering from ML kernels defined as
    memref_stream to snitch_stream + riscv.
    """

    name = "test-lower-memref-stream-to-snitch-stream"

    passes: tuple[ModulePass, ...] = TEST_LOWER_MEMREF_STREAM_TO_SNITCH_STREAM
